#!/usr/bin/env python
# -*- coding: utf-8 -*-
# flake8: noqa

import argparse
import dask
import dask.array as da
import numpy as np
from astropy.io import fits
import warnings
from africanus.model.spi.dask import fit_spi_components
iFs = np.fft.ifftshift
Fs = np.fft.fftshift

# we want to fall back to numpy if pypocketfft is not installed
# so set up functions to have the same call signatures
try:
    from pypocketfft import r2c, c2r

    def fft(x, ax, ncpu):
        return r2c(x, axes=ax, forward=True,
                   nthreads=ncpu, inorm=0)

    def ifft(y, ax, ncpu, lastsize):
        return c2r(y, axes=ax, forward=False, lastsize=lastsize,
                   nthreads=args.ncpu, inorm=2)
except:
    warnings.warn("No pypocketfft installation found. "
                  "FFT's will be performed in serial. "
                  "Install pypocketfft from "
                  "https://gitlab.mpcdf.mpg.de/mtr/pypocketfft "
                  "for optimal performance.",
                  ImportWarning)
    from numpy.fft import rfftn, irfftn
    # additional arguments will have no effect

    def fft(x, ax, ncpu):
        return rfftn(x, axes=ax)

    def ifft(y, ax, ncpu, lastsize):
        return irfftn(y, axes=ax)


def Gaussian2D(xin, yin, GaussPar=(1., 1., 0.)):
    S0, S1, PA = GaussPar
    PA = 90 + PA
    SMaj = np.max([S0, S1])
    SMin = np.min([S0, S1])
    A = np.array([[1. / SMaj ** 2, 0],
                  [0, 1. / SMin ** 2]])

    c, s, t = np.cos, np.sin, PA
    R = np.array([[c(t), -s(t)],
                  [s(t), c(t)]])
    A = np.dot(np.dot(R.T, A), R)
    sOut = xin.shape
    # only compute the result where necessary
    extent = (5 * SMaj)**2
    xflat = xin.squeeze()
    yflat = yin.squeeze()
    ind = np.argwhere(xflat**2 + yflat**2 <= extent).squeeze()
    idx = ind[:, 0]
    idy = ind[:, 1]
    x = np.array([xflat[idx, idy].ravel(), yflat[idx, idy].ravel()])
    R = np.einsum('nb,bc,cn->n', x.T, A, x)
    # need to adjust for the fact that GaussPar corresponds to FWHM
    fwhm_conv = 2*np.sqrt(2*np.log(2))
    tmp = np.exp(-fwhm_conv*R)
    gausskern = np.zeros_like(xflat, dtype=np.float64)
    gausskern[idx, idy] = tmp
    return np.ascontiguousarray(gausskern.reshape(sOut),
                                dtype=np.float64)


def convolve_model(model, gausskern, args):
    print("Doing FFT's")
    # get padding
    _, npix_l, npix_m = model.shape
    pfrac = args.padding_frac/2.0
    npad_l = int(pfrac*npix_l)
    npad_m = int(pfrac*npix_m)
    # get fast FFT sizes
    try:
        from scipy.fftpack import next_fast_len
        nfft = next_fast_len(npix_l + 2*npad_l)
        npad_ll = (nfft - npix_l)//2
        npad_lr = nfft - npix_l - npad_ll
        nfft = next_fast_len(npix_m + 2*npad_m)
        npad_ml = (nfft - npix_m)//2
        npad_mr = nfft - npix_m - npad_ml
        padding = ((0, 0), (npad_ll, npad_lr), (npad_ml, npad_mr))
        unpad_l = slice(npad_ll, -npad_lr)
        unpad_m = slice(npad_ml, -npad_mr)
    except:
        warnings.warn("Could not determine fast fft size. "
                      "Install scipy for optimal performance.",
                      ImportWarning)
        padding = ((0, 0), (npad_l, npad_l), (npad_m, npad_m))
        unpad_l = slice(npad_l, -npad_l)
        unpad_m = slice(npad_m, -npad_m)
    ax = (1, 2)  # axes over which to perform fft
    lastsize = npix_m + np.sum(padding[-1])

    # get FT of convolution kernel
    gausskernhat = fft(iFs(np.pad(gausskern[None], padding, mode='constant'),
                           axes=ax), ax, args.ncpu)

    # Convolve model with Gaussian kernel
    convmodel = fft(iFs(np.pad(model, padding, mode='constant'), axes=ax),
                    ax, args.ncpu)
    convmodel *= gausskernhat
    return Fs(ifft(convmodel, ax, args.ncpu, lastsize),
              axes=ax)[:, unpad_l, unpad_m]


def interpolate_beam(xx, yy, maskindices, freqs, args):
    print("Interpolating beam")
    l_source = xx[maskindices[:, 0], maskindices[:, 1]]
    m_source = yy[maskindices[:, 0], maskindices[:, 1]]

    lm_source = np.vstack((l_source.ravel(), m_source.ravel())).T

    ntime = 1
    nant = 1
    nband = freqs.size
    parangles = np.zeros((ntime, nant,), dtype=np.float64)
    ant_scale = np.ones((nant, nband, 2), dtype=np.float64)
    point_errs = np.zeros((ntime, nant, nband, 2), dtype=np.float64)

    if args.beammodel == "eidos":
        raise NotImplementedError("eidos is coming!!!")
    else:
        print("Loading fits beam patterns from %s" % args.beammodel)
        from glob import glob
        paths = glob(args.beammodel + '_**_**.fits')
        beam_hdr = None
        for path in paths:
            if 'xx' in path or 'XX' in path or 'rr' in path or 'RR' in path:
                if 're' in path:
                    corr1_re = fits.getdata(path)
                    if beam_hdr is None:
                        beam_hdr = fits.getheader(path)
                elif 'im' in path:
                    corr1_im = fits.getdata(path)
                else:
                    raise NotImplementedError("Only re/im patterns supported")
            elif 'yy' in path or 'YY' in path or 'll' in path or 'LL' in path:
                if 're' in path:
                    corr2_re = fits.getdata(path)
                elif 'im' in path:
                    corr2_im = fits.getdata(path)
                else:
                    raise NotImplementedError("Only re/im patterns supported")
        # get Stokes I amplitude
        beam_amp = (corr1_re**2 + corr1_im**2 + corr2_re**2 + corr2_im**2)/2.0
        # get cube in correct shape for interpolation code
        beam_amp = np.ascontiguousarray(np.transpose(beam_amp, (1, 2, 0))
                                        [:, :, :, None, None])
        # get cube info
        if beam_hdr['CUNIT1'] != "DEG" and beam_hdr['CUNIT1'] != "deg":
            raise ValueError("Beam image units must be in degrees")
        npix_l = beam_hdr['NAXIS1']
        refpix_l = beam_hdr['CRPIX1']
        delta_l = beam_hdr['CDELT1']
        l_min = (1 - refpix_l)*delta_l
        l_max = (1 + npix_l - refpix_l)*delta_l

        if beam_hdr['CUNIT2'] != "DEG" and beam_hdr['CUNIT2'] != "deg":
            raise ValueError("Beam image units must be in degrees")
        npix_m = beam_hdr['NAXIS2']
        refpix_m = beam_hdr['CRPIX2']
        delta_m = beam_hdr['CDELT2']
        m_min = (1 - refpix_m)*delta_m
        m_max = (1 + npix_m - refpix_m)*delta_m

        if (l_min > l_source.min() or m_min > m_source.min() or
                l_max < l_source.max() or m_max < m_source.max()):
            raise ValueError("The supplied beam is not large enough")

        beam_extents = np.array([[l_min, l_max], [m_min, m_max]])

        # get frequencies
        if beam_hdr["CTYPE3"] != 'FREQ':
            raise ValueError(
                "Cubes are assumed to be in format [nchan, nx, ny]")
        nchan = beam_hdr['NAXIS3']
        refpix = beam_hdr['CRPIX3']
        delta = beam_hdr['CDELT3']  # assumes units are Hz
        freq0 = beam_hdr['CRVAL3']
        bfreqs = freq0 + np.arange(1 - refpix, 1 + nchan - refpix) * delta
        if bfreqs[0] > freqs[0] or bfreqs[-1] < freqs[-1]:
            warnings.warn("The supplied beam does not have sufficient "
                          "bandwidth. Beam frequencies:")
            with np.printoptions(precision=2):
                print(bfreqs)

        # LB - dask probably not necessary for small problem
        from africanus.rime.fast_beam_cubes import beam_cube_dde
        beam_source = beam_cube_dde(beam_amp, beam_extents, bfreqs,
                                    lm_source, parangles, point_errs,
                                    ant_scale, freqs).squeeze()
        return beam_source


def create_parser():
    p = argparse.ArgumentParser(description='Simple spectral index fitting'
                                            'tool.',
                                formatter_class=argparse.RawTextHelpFormatter)
    p.add_argument("--fitsmodel", type=str, required=True)
    p.add_argument("--fitsresidual", type=str)
    p.add_argument('--outfile', type=str,
                   help="Path to output directory. \n"
                        "Placed next to input model if outfile not provided.")
    p.add_argument('--beampars', default=None, nargs='+', type=float,
                   help="Beam parameters matching FWHM of restoring beam "
                        "specified as emaj emin pa. \n"
                        "By default these are taken from the fits header "
                        "of the residual image.")
    p.add_argument('--threshold', default=5, type=float,
                   help="Multiple of the rms in the residual to threshold "
                        "on. \n"
                        "Only components above threshold*rms will be fit.")
    p.add_argument('--maxDR', default=100, type=float,
                   help="Maximum dynamic range used to determine the "
                        "threshold above which components need to be fit. \n"
                        "Only used if residual is not passed in.")
    p.add_argument('--ncpu', default=0, type=int,
                   help="Number of threads to use. \n"
                        "Default of zero means use all threads")
    p.add_argument('--beammodel', default=None, type=str,
                   help="Fits beam model to use. \n"
                        "It is assumed that the pattern is path_to_beam/"
                        "name_corr_re/im.fits. \n"
                        "Provide only the path up to name "
                        "e.g. /home/user/beams/meerkat_lband. \n"
                        "Patterns mathing corr are determined "
                        "automatically. \n"
                        "Only real and imaginary beam models currently "
                        "supported.")
    p.add_argument('--output', default='aiIbc', type=str,
                   help="Outputs to write. Letter correspond to: \n"
                   "a - alpha map \n"
                   "i - I0 map \n"
                   "I - reconstructed cube form alpha and I0 \n"
                   "b - interpolated beam \n"
                   "c - restoring beam used for convolution \n"
                   "Default is to write all of them")
    p.add_argument("--padding_frac", default=0.2, type=float,
                   help="Padding factor for FFT's.")
    return p


def main(args):

    ref_hdr = fits.getheader(args.fitsresidual)
    if args.beampars is None:
        print("Attempting to take beampars from residual fits header")
        emaj = ref_hdr['BMAJ1']
        emin = ref_hdr['BMIN1']
        pa = ref_hdr['BPA1']
        beampars = (emaj, emin, pa)
    else:
        beampars = tuple(args.beampars)
        # emaj, emin, pa = args.beampars
    print("Using emaj = %3.2e, emin = %3.2e, PA = %3.2e" % beampars)

    # load images
    model = np.ascontiguousarray(fits.getdata(args.fitsmodel).squeeze(),
                                 dtype=np.float64)
    mhdr = fits.getheader(args.fitsmodel)

    if mhdr['CUNIT1'] != "DEG" and mhdr['CUNIT1'] != "deg":
        raise ValueError("Image units must be in degrees")
    npix_l = mhdr['NAXIS1']
    refpix_l = mhdr['CRPIX1']
    delta_l = mhdr['CDELT1']
    l_coord = np.arange(1 - refpix_l, 1 + npix_l - refpix_l)*delta_l

    if mhdr['CUNIT2'] != "DEG" and mhdr['CUNIT2'] != "deg":
        raise ValueError("Image units must be in degrees")
    npix_m = mhdr['NAXIS2']
    refpix_m = mhdr['CRPIX2']
    delta_m = mhdr['CDELT2']
    m_coord = np.arange(1 - refpix_m, 1 + npix_m - refpix_m)*delta_m

    print("Image shape = ", (npix_l, npix_m))

    # get frequencies
    if mhdr["CTYPE4"] == 'FREQ':
        freq_axis = 4
        nband = mhdr['NAXIS4']
        refpix_nu = mhdr['CRPIX4']
        delta_nu = mhdr['CDELT4']  # assumes units are Hz
        ref_freq = mhdr['CRVAL4']
        ncorr = mhdr['NAXIS3']
    elif mhdr["CTYPE3"] == 'FREQ':
        freq_axis = 3
        nband = mhdr['NAXIS3']
        refpix_nu = mhdr['CRPIX3']
        delta_nu = mhdr['CDELT3']  # assumes units are Hz
        ref_freq = mhdr['CRVAL3']
        ncorr = mhdr['NAXIS4']
    else:
        raise ValueError("Freq axis must be 3rd or 4th")

    if ncorr > 1:
        raise ValueError("Only Stokes I cubes supported")

    freqs = ref_freq + np.arange(1 - refpix_nu,
                                 1 + nband - refpix_nu) * delta_nu

    print("Cube frequencies:")
    with np.printoptions(precision=2):
        print(freqs)
    print("Reference frequency is %3.2e Hz " % ref_freq)

    # get the Gaussian convolution kernel
    xx, yy = np.meshgrid(l_coord, m_coord)
    gausskern = Gaussian2D(xx, yy, beampars)

    # Convolve model with Gaussian restroring beam at lowest frequency
    model = convolve_model(model, gausskern, args)

    # set threshold
    if args.fitsresidual is not None:
        resid = fits.getdata(args.fitsresidual).squeeze().astype(np.float64)
        rms = np.std(resid)
        rms_cube = np.std(resid.reshape(nband, npix_l*npix_m), axis=1).ravel()
        threshold = args.threshold * rms
        print("Setting cutoff threshold as %i times the rms "
              "of the residual" % args.threshold)
        del resid
    else:
        print("No residual provided. Setting  threshold i.t.o dynamic range. "
              "Max dynamic range is %i" % args.maxDR)
        threshold = model.max()/args.maxDR
        if args.channelweights is None:
            rms_cube = None

    print("Threshold set to %f Jy." % threshold)

    # get pixels above threshold
    minimage = np.amin(model, axis=0)
    maskindices = np.argwhere(minimage > threshold)
    if not maskindices.size:
        raise ValueError("No components found above threshold. "
                         "Try lowering your threshold."
                         "Max of convolved model is %3.2e" % model.max())
    fitcube = model[:, maskindices[:, 0], maskindices[:, 1]].T

    print(xx.shape, yy.shape, maskindices.shape)

    # get primary beam at source locations
    if args.beammodel is not None:
        beam_source = interpolate_beam(xx, yy, maskindices, freqs, args)
        # correct cube
        fitcube /= beam_source

    # set weights for fit
    if rms_cube is not None:
        print("Using RMS in each imaging band to determine weights.")
        weights = np.where(rms_cube > 0, 1.0/rms_cube**2, 0.0)
        # normalise
        weights /= weights.max()
    else:
        print("No residual provided. Using equal weights.")
        weights = np.ones(nband, dtype=np.float64)

    ncomps, _ = fitcube.shape
    fitcube = da.from_array(fitcube.astype(np.float64),
                            chunks=(ncomps//args.ncpu, nband))
    weights = da.from_array(weights.astype(np.float64), chunks=(nband))
    freqsdask = da.from_array(freqs.astype(np.float64), chunks=(nband))

    print("Fitting %i components" % ncomps)
    alpha, _, Iref, _ = fit_spi_components(fitcube, weights, freqsdask,
                                           np.float64(ref_freq)).compute()
    print("Done. Writing output.")

    alphamap = np.zeros([npix_l, npix_m])
    i0map = np.zeros([npix_l, npix_m])
    alphamap[maskindices[:, 0], maskindices[:, 1]] = alpha
    i0map[maskindices[:, 0], maskindices[:, 1]] = Iref

    # save next to model if no outfile is provided
    if args.outfile is None:
        # find last /
        tmp = args.fitsmodel[::-1]
        idx = tmp.find('/')
        if idx != -1:
            outfile = args.fitsmodel[0:-idx]
        else:
            outfile = 'image-'
    else:
        outfile = args.outfile

    hdu = fits.PrimaryHDU(header=mhdr)
    if 'I' in args.output:
        # get the reconstructed cube
        Irec_cube = i0map[None, :, :] * \
            (freqs[:, None, None]/ref_freq)**alphamap[None, :, :]
        # save it
        if freq_axis == 3:
            hdu.data = Irec_cube[None, :, :, :]
        elif freq_axis == 4:
            hdu.data = Irec_cube[:, None, :, :]
        name = outfile + 'Irec_cube.fits'
        hdu.writeto(name, overwrite=True)
        print("Wrote reconstructed cube to %s" % name)

    if args.beammodel is not None and 'b' in args.output:
        beam_map = np.zeros((nband, npix_l, npix_m))
        beam_map[:, maskindices[:, 0], maskindices[:, 1]] = beam_source.T
        if freq_axis == 3:
            hdu.data = beam_map[None, :, :, :]
        elif freq_axis == 4:
            hdu.data = beam_map[:, None, :, :]
        name = outfile + 'interpolated_beam_cube.fits'
        hdu.writeto(name, overwrite=True)
        print("Wrote interpolated beam cube to %s" % name)

    hdr_keys = ['SIMPLE', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2', 'NAXIS3',
                'NAXIS4', 'BUNIT', 'BMAJ', 'BMIN', 'BPA', 'EQUINOX', 'BTYPE',
                'TELESCOP', 'OBSERVER', 'OBJECT', 'ORIGIN', 'CTYPE1', 'CTYPE2',
                'CTYPE3', 'CTYPE4', 'CRPIX1', 'CRPIX2', 'CRPIX3', 'CRPIX4',
                'CRVAL1', 'CRVAL2', 'CRVAL3', 'CRVAL4', 'CDELT1', 'CDELT2',
                'CDELT3', 'CDELT4', 'CUNIT1', 'CUNIT2', 'CUNIT3', 'CUNIT4',
                'SPECSYS', 'DATE-OBS']

    new_hdr = {}
    for key in hdr_keys:
        new_hdr[key] = ref_hdr[key]

    if freq_axis == 3:
        new_hdr["NAXIS3"] = 1
        new_hdr["CRVAL3"] = ref_freq
    elif freq_axis == 4:
        new_hdr["NAXIS4"] = 1
        new_hdr["CRVAL4"] = ref_freq

    new_hdr = fits.Header(new_hdr)

    # save alpha map
    if 'a' in args.output:
        hdu = fits.PrimaryHDU(header=new_hdr)
        hdu.data = alphamap
        name = outfile + 'alpha.fits'
        hdu.writeto(name, overwrite=True)
        print("Wrote alpha map to %s" % name)

    # save I0 map
    if 'i' in args.output:
        hdu = fits.PrimaryHDU(header=new_hdr)
        hdu.data = i0map
        name = outfile + 'I0.fits'
        hdu.writeto(name, overwrite=True)
        print("Wrote I0 map to %s" % name)

    # save clean beam for consistency check
    if 'c' in args.output:
        hdu = fits.PrimaryHDU(header=new_hdr)
        hdu.data = gausskern
        name = outfile + 'clean-beam.fits'
        hdu.writeto(name, overwrite=True)
        print("Wrote clean beam to %s" % name)

    print("All done here")


if __name__ == "__main__":
    args = create_parser().parse_args()

    if args.ncpu:
        from multiprocessing.pool import ThreadPool
        dask.config.set(pool=ThreadPool(args.ncpu))
    else:
        import multiprocessing
        args.ncpu = multiprocessing.cpu_count()

    print("Using %i threads" % args.ncpu)

    main(args)



# -*- coding: utf-8 -*-


from functools import reduce
import logging
from operator import mul
from os.path import join as pjoin

import numpy as np

from africanus.rime.predict import (PREDICT_DOCS, predict_checks)
from africanus.util.code import format_code, memoize_on_key
from africanus.util.cuda import cuda_type, grids
from africanus.util.jinja2 import jinja_env
from africanus.util.requirements import requires_optional

try:
    import cupy as cp
    from cupy.cuda.compiler import CompileException
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None


log = logging.getLogger(__name__)


_TEMPLATE_PATH = pjoin("rime", "cuda", "predict.cu.j2")


def _key_fn(*args):
    """ Hash on array datatypes and rank """
    return tuple((a.dtype, a.ndim)
                 if isinstance(a, (np.ndarray, cp.ndarray))
                 else a for a in args)


@memoize_on_key(_key_fn)
def _generate_kernel(time_index, antenna1, antenna2,
                     dde1_jones, source_coh, dde2_jones,
                     die1_jones, base_vis, die2_jones,
                     corrs, out_ndim):

    tup = predict_checks(time_index, antenna1, antenna2,
                         dde1_jones, source_coh, dde2_jones,
                         die1_jones, base_vis, die2_jones)

    (have_ddes1, have_coh, have_ddes2, have_dies1, have_bvis, have_dies2) = tup

    # Check types
    if time_index.dtype != np.int32:
        raise TypeError("time_index.dtype != np.int32 '%s'" % time_index.dtype)

    if antenna1.dtype != np.int32:
        raise TypeError("antenna1.dtype != np.int32 '%s'" % antenna1.dtype)

    if antenna2.dtype != np.int32:
        raise TypeError("antenna2.dtype != np.int32 '%s'" % antenna2.dtype)

    # Create template
    render = jinja_env.get_template(_TEMPLATE_PATH).render
    name = "predict_vis"

    # Complex output type
    out_dtype = np.result_type(dde1_jones, source_coh, dde2_jones,
                               die1_jones, base_vis, die2_jones)

    ncorrs = reduce(mul, corrs, 1)

    # corrs x channels, rows
    blockdimx = 32
    blockdimy = 24 if out_dtype == np.complex128 else 32

    block = (blockdimx, blockdimy, 1)

    code = render(kernel_name=name, blockdimx=blockdimx, blockdimy=blockdimy,
                  have_dde1=have_ddes1,
                  dde1_type=cuda_type(dde1_jones) if have_ddes1 else "int",
                  dde1_ndim=dde1_jones.ndim if have_ddes1 else 1,
                  have_dde2=have_ddes2,
                  dde2_type=cuda_type(dde2_jones) if have_ddes2 else "int",
                  dde2_ndim=dde2_jones.ndim if have_ddes2 else 1,
                  have_coh=have_coh,
                  coh_type=cuda_type(source_coh) if have_coh else "int",
                  coh_ndim=source_coh.ndim if have_coh else 1,
                  have_die1=have_dies1,
                  die1_type=cuda_type(die1_jones) if have_dies1 else "int",
                  die1_ndim=die1_jones.ndim if have_dies1 else 1,
                  have_base_vis=have_bvis,
                  base_vis_type=cuda_type(base_vis) if have_bvis else "int",
                  base_vis_ndim=base_vis.ndim if have_bvis else 1,
                  have_die2=have_dies2,
                  die2_type=cuda_type(die2_jones) if have_dies2 else "int",
                  die2_ndim=die2_jones.ndim if have_dies2 else 1,
                  out_type=cuda_type(out_dtype),
                  corrs=ncorrs,
                  out_ndim=out_ndim,
                  warp_size=32).encode('utf-8')

    return cp.RawKernel(code, name), block, out_dtype


@requires_optional("cupy", opt_import_error)
def predict_vis(time_index, antenna1, antenna2,
                dde1_jones=None, source_coh=None, dde2_jones=None,
                die1_jones=None, base_vis=None, die2_jones=None):
    """ Cupy implementation of the feed_rotation kernel. """

    have_ddes = dde1_jones is not None and dde2_jones is not None
    have_dies = die1_jones is not None and die2_jones is not None
    have_coh = source_coh is not None
    have_bvis = base_vis is not None

    # Infer the output shape
    if have_ddes:
        row = time_index.shape[0]
        chan = dde1_jones.shape[3]
        corrs = dde1_jones.shape[4:]
    elif have_coh:
        row = time_index.shape[0]
        chan = source_coh.shape[2]
        corrs = source_coh.shape[3:]
    elif have_dies:
        row = time_index.shape[0]
        chan = die1_jones.shape[2]
        corrs = die1_jones.shape[3:]
    elif have_bvis:
        row = time_index.shape[0]
        chan = base_vis.shape[1]
        corrs = base_vis.shape[2:]
    else:
        raise ValueError("Insufficient inputs supplied for determining "
                         "the output shape")

    ncorrs = len(corrs)

    # Flatten correlations
    if ncorrs == 2:
        flat_corrs = reduce(mul, corrs, 1)

        if have_ddes:
            dde_shape = dde1_jones.shape[:-ncorrs] + (flat_corrs,)
            dde1_jones = dde1_jones.reshape(dde_shape)
            dde2_jones = dde2_jones.reshape(dde_shape)

        if have_coh:
            coh_shape = source_coh.shape[:-ncorrs] + (flat_corrs,)
            source_coh = source_coh.reshape(coh_shape)

        if have_dies:
            die_shape = die1_jones.shape[:-ncorrs] + (flat_corrs,)
            die1_jones = die1_jones.reshape(die_shape)
            die2_jones = die2_jones.reshape(die_shape)

        if have_bvis:
            bvis_shape = base_vis.shape[:-ncorrs] + (flat_corrs,)
            base_vis = base_vis.reshape(bvis_shape)

    elif ncorrs == 1:
        flat_corrs = corrs[0]
    else:
        raise ValueError("Invalid correlation setup %s" % (corrs,))

    out_shape = (row, chan) + (flat_corrs,)

    kernel, block, out_dtype = _generate_kernel(time_index,
                                                antenna1,
                                                antenna2,
                                                dde1_jones,
                                                source_coh,
                                                dde2_jones,
                                                die1_jones,
                                                base_vis,
                                                die2_jones,
                                                corrs,
                                                len(out_shape))

    grid = grids((chan*flat_corrs, row, 1), block)
    out = cp.empty(shape=out_shape, dtype=out_dtype)

    # Normalise the time index
    # TODO(sjperkins)
    # Normalise the time index with a device-wide reduction
    norm_time_index = time_index - time_index.min()

    args = (norm_time_index, antenna1, antenna2,
            dde1_jones, source_coh, dde2_jones,
            die1_jones, base_vis, die2_jones,
            out)

    try:
        kernel(grid, block, tuple(a for a in args if a is not None))
    except CompileException:
        log.exception(format_code(kernel.code))
        raise

    return out.reshape((row, chan) + corrs)


try:
    predict_vis.__doc__ = PREDICT_DOCS.substitute(
                                array_type=":class:`cupy.ndarray`",
                                get_time_index=":code:`cp.unique(time, "
                                               "return_inverse=True)[1]`",
                                extra_args="",
                                extra_notes="")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


from africanus.rime.phase import (phase_delay as np_phase_delay,
                                  PHASE_DELAY_DOCS)
from africanus.rime.parangles import parallactic_angles as np_parangles
from africanus.rime.feeds import feed_rotation as np_feed_rotation
from africanus.rime.feeds import FEED_ROTATION_DOCS
from africanus.rime.transform import transform_sources as np_transform_sources
from africanus.rime.fast_beam_cubes import (beam_cube_dde as np_beam_cube_dde,
                                            BEAM_CUBE_DOCS)
from africanus.rime.dask_predict import predict_vis  # noqa
from africanus.rime.zernike import zernike_dde as np_zernike_dde


from africanus.util.docs import mod_docs
from africanus.util.requirements import requires_optional
from africanus.util.type_inference import infer_complex_dtype

import numpy as np

try:
    import dask.array as da
except ImportError as e:
    da_import_error = e
else:
    da_import_error = None


def _phase_delay_wrap(lm, uvw, frequency):
    return np_phase_delay(lm[0], uvw[0], frequency)


@requires_optional('dask.array', da_import_error)
def phase_delay(lm, uvw, frequency):
    """ Dask wrapper for phase_delay function """
    return da.core.blockwise(_phase_delay_wrap, ("source", "row", "chan"),
                             lm, ("source", "(l,m)"),
                             uvw, ("row", "(u,v,w)"),
                             frequency, ("chan",),
                             dtype=infer_complex_dtype(lm, uvw, frequency))


def _parangle_wrapper(t, ap, fc, **kw):
    return np_parangles(t, ap[0], fc[0], **kw)


@requires_optional('dask.array', da_import_error)
def parallactic_angles(times, antenna_positions, field_centre, **kwargs):

    return da.core.blockwise(_parangle_wrapper, ("time", "ant"),
                             times, ("time",),
                             antenna_positions, ("ant", "xyz"),
                             field_centre, ("fc",),
                             dtype=times.dtype,
                             **kwargs)


@requires_optional('dask.array', da_import_error)
def feed_rotation(parallactic_angles, feed_type):
    pa_dims = tuple("pa-%d" % i for i in range(parallactic_angles.ndim))
    corr_dims = ('corr-1', 'corr-2')

    if parallactic_angles.dtype == np.float32:
        dtype = np.complex64
    elif parallactic_angles.dtype == np.float64:
        dtype = np.complex128
    else:
        raise ValueError("parallactic_angles have "
                         "non-floating point dtype")

    return da.core.blockwise(np_feed_rotation, pa_dims + corr_dims,
                             parallactic_angles, pa_dims,
                             feed_type=feed_type,
                             new_axes={'corr-1': 2, 'corr-2': 2},
                             dtype=dtype)


def _xform_wrap(lm, parallactic_angles, pointing_errors,
                antenna_scaling, frequency, dtype_):
    return np_transform_sources(lm[0], parallactic_angles,
                                pointing_errors[0], antenna_scaling,
                                frequency, dtype=dtype_)


@requires_optional('dask.array', da_import_error)
def transform_sources(lm, parallactic_angles, pointing_errors,
                      antenna_scaling, frequency, dtype=None):

    if dtype is None:
        dtype = np.float64

    xform_inds = ("comp", "src", "time", "ant", "chan")

    return da.core.blockwise(_xform_wrap, xform_inds,
                             lm, ("src", "lm"),
                             parallactic_angles, ("time", "ant"),
                             pointing_errors, ("time", "ant", "lm"),
                             antenna_scaling, ("ant", "chan"),
                             frequency, ("chan",),
                             new_axes={"comp": 3},
                             dtype=dtype,
                             dtype_=dtype)


def _beam_cube_dde_wrapper(beam, beam_lm_extents, beam_freq_map,
                           lm, parallactic_angles,
                           point_errors, antenna_scaling,
                           frequencies):
    return np_beam_cube_dde(beam[0][0][0], beam_lm_extents[0][0],
                            beam_freq_map[0], lm[0],
                            parallactic_angles, point_errors[0],
                            antenna_scaling[0], frequencies)


@requires_optional('dask.array', da_import_error)
def beam_cube_dde(beam, beam_lm_extents, beam_freq_map,
                  lm, parallactic_angles,
                  point_errors, antenna_scaling,
                  frequencies):

    if not all(len(c) == 1 for c in beam.chunks):
        raise ValueError("Beam chunking unsupported")

    if not all(len(c) == 1 for c in beam_freq_map.chunks):
        raise ValueError("Beam frequency map chunking unsupported")

    if not all(len(c) == 1 for c in beam_lm_extents.chunks):
        raise ValueError("Chunking of beam_lm_extents unsupported")

    corr_shapes = beam.shape[3:]
    corr_dims = tuple("corr-%d" % i for i in range(len(corr_shapes)))

    dde_dims = ("source", "time", "ant", "chan") + corr_dims
    beam_dims = ("beam-lw", "beam-mh", "beam-nud") + corr_dims

    return da.core.blockwise(_beam_cube_dde_wrapper, dde_dims,
                             beam, beam_dims,
                             beam_lm_extents, ("beam-lm", "beam-ext"),
                             beam_freq_map, ("beam-nud",),
                             lm, ("source", "source-comp"),
                             parallactic_angles, ("time", "ant"),
                             point_errors, ("time", "ant", "chan", "pt-comp"),
                             antenna_scaling, ("ant", "chan", "scale-comp"),
                             frequencies, ("chan",),
                             dtype=beam.dtype)


def _zernike_wrapper(coords, coeffs, noll_index):
    # coords loses "three" dim
    # coeffs loses "poly" dim
    # noll_index loses "poly" dim
    return np_zernike_dde(coords[0], coeffs[0], noll_index[0])


@requires_optional('dask.array', da_import_error)
def zernike_dde(coords, coeffs, noll_index):
    ncorrs = len(coeffs.shape[2:-1])
    corr_dims = tuple("corr-%d" % i for i in range(ncorrs))

    return da.core.blockwise(_zernike_wrapper,
                             ("source", "time", "ant", "chan") + corr_dims,
                             coords,
                             ("three", "source", "time", "ant", "chan"),
                             coeffs,
                             ("ant", "chan") + corr_dims + ("poly",),
                             noll_index,
                             ("ant", "chan") + corr_dims + ("poly",),
                             dtype=coeffs.dtype)


try:
    phase_delay.__doc__ = PHASE_DELAY_DOCS.substitute(
                            array_type=":class:`dask.array.Array`")
except AttributeError:
    pass

try:
    parallactic_angles.__doc__ = mod_docs(np_parangles.__doc__,
                                          [(":class:`numpy.ndarray`",
                                            ":class:`dask.array.Array`")])
except AttributeError:
    pass

try:
    feed_rotation.__doc__ = FEED_ROTATION_DOCS.substitute(
                                array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass

try:
    transform_sources.__doc__ = mod_docs(np_transform_sources.__doc__,
                                         [(":class:`numpy.ndarray`",
                                           ":class:`dask.array.Array`")])
except AttributeError:
    pass

try:
    beam_cube_dde.__doc__ = BEAM_CUBE_DOCS.substitute(
                                array_type=":class:`dask.array.Array`")
except AttributeError:
    pass

try:
    zernike_dde.__doc__ = mod_docs(np_zernike_dde.__doc__,
                                   [(":class:`numpy.ndarray`",
                                     ":class:`dask.array.Array`")])
except AttributeError:
    pass



# -*- coding: utf-8 -*-


from africanus.util.testing import mark_in_pytest


# content of conftest.py
def pytest_configure(config):
    mark_in_pytest(True)


def pytest_unconfigure(config):
    mark_in_pytest(False)



# -*- coding: utf-8 -*-


import math

import numpy as np

from africanus.constants import minus_two_pi_over_c
from africanus.util.docs import DocstringTemplate
from africanus.util.numba import generated_jit
from africanus.util.type_inference import infer_complex_dtype


@generated_jit(nopython=True, nogil=True, cache=True)
def phase_delay(lm, uvw, frequency):
    # Bake constants in with the correct type
    one = lm.dtype(1.0)
    neg_two_pi_over_c = lm.dtype(minus_two_pi_over_c)

    out_dtype = infer_complex_dtype(lm, uvw, frequency)

    def _phase_delay_impl(lm, uvw, frequency):
        shape = (lm.shape[0], uvw.shape[0], frequency.shape[0])
        complex_phase = np.zeros(shape, dtype=out_dtype)

        # For each source
        for source in range(lm.shape[0]):
            l, m = lm[source]
            n = math.sqrt(one - l**2 - m**2) - one

            # For each uvw coordinate
            for row in range(uvw.shape[0]):
                u, v, w = uvw[row]
                # e^(-2*pi*(l*u + m*v + n*w)/c)
                real_phase = neg_two_pi_over_c * (l * u + m * v + n * w)

                # Multiple in frequency for each channel
                for chan in range(frequency.shape[0]):
                    p = real_phase * frequency[chan]

                    # Our phase input is purely imaginary
                    # so we can can elide a call to exp
                    # and just compute the cos and sin
                    complex_phase.real[source, row, chan] = math.cos(p)
                    complex_phase.imag[source, row, chan] = math.sin(p)

        return complex_phase

    return _phase_delay_impl


PHASE_DELAY_DOCS = DocstringTemplate(
    r"""
    Computes the phase delay (K) term:

    .. math::

        & {\Large e^{-2 \pi i (u l + v m + w (n - 1))} }

        & \textrm{where } n = \sqrt{1 - l^2 - m^2}

    Notes
    -----

    Corresponds to the complex exponential of the `Van Cittert-Zernike Theorem
    <https://en.wikipedia.org/wiki/Van_Cittert%E2%80%93Zernike_theorem_>`_.

    `MeqTrees
    <https://github.com/ska-sa/meqtrees-timba/blob/
    6a7e873d4d1fe538981dec5851418cbd371b8388/MeqNodes/src/PSVTensor.cc#L314_>`_
    uses a positive sign convention and so any UVW coordinates must be inverted
    in order for their phase delay terms (and therefore visibilities) to agree.

    Parameters
    ----------

    lm : $(array_type)
        LM coordinates of shape :code:`(source, 2)` with
        L and M components in the last dimension.
    uvw : $(array_type)
        UVW coordinates of shape :code:`(row, 3)` with
        U, V and W components in the last dimension.
    frequency : $(array_type)
        frequencies of shape :code:`(chan,)`

    Returns
    -------
    complex_phase : $(array_type)
        complex of shape :code:`(source, row, chan)`
    """)

try:
    phase_delay.__doc__ = PHASE_DELAY_DOCS.substitute(
                            array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


from functools import reduce
from operator import mul

import numpy as np

from africanus.util.docs import DocstringTemplate
from africanus.util.numba import jit


@jit(nopython=True, nogil=True, cache=True)
def _nb_feed_rotation(parallactic_angles, feed_type, feed_rotation):
    shape = parallactic_angles.shape
    parangles = parallactic_angles.flat

    # Linear feeds
    if feed_type == 0:
        for i, pa in enumerate(parangles):
            pa_cos = np.cos(pa)
            pa_sin = np.sin(pa)

            feed_rotation.real[i, 0, 0] = pa_cos
            feed_rotation.imag[i, 0, 0] = 0.0
            feed_rotation.real[i, 0, 1] = pa_sin
            feed_rotation.imag[i, 0, 1] = 0.0
            feed_rotation.real[i, 1, 0] = -pa_sin
            feed_rotation.imag[i, 1, 0] = 0.0
            feed_rotation.real[i, 1, 1] = pa_cos
            feed_rotation.imag[i, 1, 1] = 0.0

    # Circular feeds
    elif feed_type == 1:
        for i, pa in enumerate(parangles):
            pa_cos = np.cos(pa)
            pa_sin = np.sin(pa)

            feed_rotation.real[i, 0, 0] = pa_cos
            feed_rotation.imag[i, 0, 0] = -pa_sin
            feed_rotation[i, 0, 1] = 0.0 + 0.0*1j
            feed_rotation[i, 1, 0] = 0.0 + 0.0*1j
            feed_rotation.real[i, 1, 1] = pa_cos
            feed_rotation.imag[i, 1, 1] = pa_sin
    else:
        raise ValueError("Invalid feed_type")

    return feed_rotation.reshape(shape + (2, 2))


def feed_rotation(parallactic_angles, feed_type='linear'):
    if feed_type == 'linear':
        poltype = 0
    elif feed_type == 'circular':
        poltype = 1
    else:
        raise ValueError("Invalid feed_type '%s'" % feed_type)

    if parallactic_angles.dtype == np.float32:
        dtype = np.complex64
    elif parallactic_angles.dtype == np.float64:
        dtype = np.complex128
    else:
        raise ValueError("parallactic_angles has "
                         "none-floating point type %s"
                         % parallactic_angles.dtype)

    # Create result array with flattened parangles
    shape = (reduce(mul, parallactic_angles.shape),) + (2, 2)
    result = np.empty(shape, dtype=dtype)

    return _nb_feed_rotation(parallactic_angles, poltype, result)


FEED_ROTATION_DOCS = DocstringTemplate(r"""
Computes the 2x2 feed rotation (L) matrix
from the ``parallactic_angles``.

.. math::

    \textrm{linear}
    \begin{bmatrix}
    cos(pa) & sin(pa) \\
    -sin(pa) & cos(pa)
    \end{bmatrix}
    \qquad
    \textrm{circular}
    \begin{bmatrix}
    e^{-i pa} & 0 \\
    0 & e^{i pa}
    \end{bmatrix}

Parameters
----------
parallactic_angles : $(array_type)
    floating point parallactic angles. Of shape
    :code:`(pa0, pa1, ..., pan)`.
feed_type : {'linear', 'circular'}
    The type of feed

Returns
-------
feed_matrix : $(array_type)
    Feed rotation matrix of shape :code:`(pa0, pa1,...,pan,2,2)`
""")

try:
    feed_rotation.__doc__ = FEED_ROTATION_DOCS.substitute(
                                array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


import numpy as np


def kron_matvec(A, b):
    """
    Computes the matrix vector product of
    a kronecker matrix in linear time.

    Parameters
    ----------
    A : :class:`numpy.ndarray`
        An array of arrays holding
        matrices [K0, K1, ...] where
        :math:`A = K_0 \\otimes K_1 \\otimes \\cdots`
    b : :class:`numpy.ndarray`
        The right hand side vector

    Returns
    -------
    x : :class:`numpy.ndarray`
        The result of :code:`A.dot(b)`
    """
    D = A.shape[0]
    N = b.size
    x = b
    for d in range(D):
        Gd = A[d].shape[0]
        X = np.reshape(x, (Gd, N//Gd))
        Z = np.einsum("ab,bc->ac", A[d], X)
        Z = np.einsum("ab -> ba", Z)
        x = Z.flatten()
    return x


def kron_cholesky(A):
    """
    Computes the Cholesky decomposition
    of a kronecker matrix as a kronecker
    matrix of Cholesky factors.

    Parameters
    ----------
    A : :class:`numpy.ndarray`
        An array of arrays holding
        matrices [K0, K1, ...] where
        :math:`A = K_0 \\otimes K_1 \\otimes \\cdots`

    Returns
    -------
    L : :class:`numpy.ndarray`
        An array of arrays holding
        matrices [L0, L1, ...] where
        :math:`L = L_0 \\otimes L_1 \\otimes \\cdots`
        and each :code:`Li = cholesky(Ki)`
    """
    D = A.shape[0]
    L = np.zeros_like(A)
    for i in range(D):
        try:
            L[i] = np.linalg.cholesky(A[i])
        except Exception:  # add jitter
            L[i] = np.linalg.cholesky(A[i] + 1e-13*np.eye(A[i].shape[0]))
    return L



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `codex-africanus` package."""

import numpy as np
from numpy.testing import assert_array_almost_equal

import pytest


def rf(*a, **kw):
    return np.random.random(*a, **kw)


def rc(*a, **kw):
    return rf(*a, **kw) + 1j*rf(*a, **kw)


chunk_parametrization = pytest.mark.parametrize("chunks", [
    {
        'source':  (2, 3, 4, 2, 2, 2, 2, 2, 2),
        'time': (2, 1, 1),
        'rows': (4, 4, 2),
        'antenna': (4,),
        'channels': (3, 2),
    }])

corr_shape_parametrization = pytest.mark.parametrize(
    'corr_shape, idm, einsum_sig1, einsum_sig2', [
        ((1,), (1,), "srci,srci,srci->rci", "rci,rci,rci->rci"),
        ((2,), (1, 1), "srci,srci,srci->rci", "rci,rci,rci->rci"),
        ((2, 2), ((1, 0), (0, 1)),
            "srcij,srcjk,srclk->rcil", "rcij,rcjk,rclk->rcil")
    ])


dde_presence_parametrization = pytest.mark.parametrize('a1j,blj,a2j', [
    [True, True, True],
    [True, False, True],
    [False, True, False],
])

die_presence_parametrization = pytest.mark.parametrize('g1j,bvis,g2j', [
    [True, True, True],
    [True, False, True],
    [False, True, False],
])


@corr_shape_parametrization
@dde_presence_parametrization
@die_presence_parametrization
@chunk_parametrization
def test_predict_vis(corr_shape, idm, einsum_sig1, einsum_sig2,
                     a1j, blj, a2j, g1j, bvis, g2j,
                     chunks):
    from africanus.rime.predict import predict_vis

    s = sum(chunks['source'])
    t = sum(chunks['time'])
    a = sum(chunks['antenna'])
    c = sum(chunks['channels'])
    r = sum(chunks['rows'])

    a1_jones = rc((s, t, a, c) + corr_shape)
    bl_jones = rc((s, r, c) + corr_shape)
    a2_jones = rc((s, t, a, c) + corr_shape)
    g1_jones = rc((t, a, c) + corr_shape)
    base_vis = rc((r, c) + corr_shape)
    g2_jones = rc((t, a, c) + corr_shape)

    #  Row indices into the above time/ant indexed arrays
    time_idx = np.asarray([0, 0, 1, 1, 2, 2, 2, 2, 3, 3])
    ant1 = np.asarray([0, 0, 0, 0, 1, 1, 1, 2, 2, 3])
    ant2 = np.asarray([0, 1, 2, 3, 1, 2, 3, 2, 3, 3])

    assert ant1.size == r

    model_vis = predict_vis(time_idx, ant1, ant2,
                            a1_jones if a1j else None,
                            bl_jones if blj else None,
                            a2_jones if a2j else None,
                            g1_jones if g1j else None,
                            base_vis if bvis else None,
                            g2_jones if g2j else None)

    assert model_vis.shape == (r, c) + corr_shape

    def _id(array):
        return np.broadcast_to(idm, array.shape)

    # For einsum, convert (time, ant) dimensions to row
    # or ID matrices if the input is present
    a1_jones = a1_jones[:, time_idx, ant1] if a1j else _id(bl_jones)
    bl_jones = bl_jones if blj else _id(bl_jones)
    a2_jones = a2_jones[:, time_idx, ant2].conj() if a2j else _id(bl_jones)

    v = np.einsum(einsum_sig1, a1_jones, bl_jones, a2_jones)

    if bvis:
        v += base_vis

    # Convert (time, ant) dimensions to row or
    # or ID matrices if input is not present
    g1_jones = g1_jones[time_idx, ant1] if g1j else _id(v)
    g2_jones = g2_jones[time_idx, ant2].conj() if g2j else _id(v)

    v = np.einsum(einsum_sig2, g1_jones, v, g2_jones)

    assert_array_almost_equal(v, model_vis)


@corr_shape_parametrization
@dde_presence_parametrization
@die_presence_parametrization
@chunk_parametrization
@pytest.mark.parametrize("streams", [1, 3])
def test_dask_predict_vis(corr_shape, idm, einsum_sig1, einsum_sig2,
                          a1j, blj, a2j, g1j, bvis, g2j,
                          chunks, streams):

    da = pytest.importorskip('dask.array')
    import numpy as np
    import dask

    from africanus.rime.predict import predict_vis as np_predict_vis
    from africanus.rime.dask import predict_vis

    # chunk sizes
    sc = chunks['source']
    tc = chunks['time']
    rrc = chunks['rows']
    ac = chunks['antenna']
    cc = chunks['channels']

    # dimension sizes
    s = sum(sc)       # sources
    t = sum(tc)       # times
    a = sum(ac)       # antennas
    c = sum(cc)       # channels
    r = sum(rrc)      # rows

    a1_jones = rc((s, t, a, c) + corr_shape)
    a2_jones = rc((s, t, a, c) + corr_shape)
    bl_jones = rc((s, r, c) + corr_shape)
    g1_jones = rc((t, a, c) + corr_shape)
    base_vis = rc((r, c) + corr_shape)
    g2_jones = rc((t, a, c) + corr_shape)

    #  Row indices into the above time/ant indexed arrays
    time_idx = np.asarray([0, 0, 1, 1, 2, 2, 2, 2, 3, 3])
    ant1 = np.asarray([0, 0, 0, 0, 1, 1, 1, 2, 2, 3])
    ant2 = np.asarray([0, 1, 2, 3, 1, 2, 3, 2, 3, 3])

    assert ant1.size == r

    np_model_vis = np_predict_vis(time_idx, ant1, ant2,
                                  a1_jones if a1j else None,
                                  bl_jones if blj else None,
                                  a2_jones if a2j else None,
                                  g1_jones if g1j else None,
                                  base_vis if bvis else None,
                                  g2_jones if g2j else None)

    da_time_idx = da.from_array(time_idx, chunks=rrc)
    da_ant1 = da.from_array(ant1, chunks=rrc)
    da_ant2 = da.from_array(ant2, chunks=rrc)

    da_a1_jones = da.from_array(a1_jones, chunks=(sc, tc, ac, cc) + corr_shape)
    da_bl_jones = da.from_array(bl_jones, chunks=(sc, rrc, cc) + corr_shape)
    da_a2_jones = da.from_array(a2_jones, chunks=(sc, tc, ac, cc) + corr_shape)
    da_g1_jones = da.from_array(g1_jones, chunks=(tc, ac, cc) + corr_shape)
    da_base_vis = da.from_array(base_vis, chunks=(rrc, cc) + corr_shape)
    da_g2_jones = da.from_array(g2_jones, chunks=(tc, ac, cc) + corr_shape)

    args = (da_time_idx, da_ant1, da_ant2,
            da_a1_jones if a1j else None,
            da_bl_jones if blj else None,
            da_a2_jones if a2j else None,
            da_g1_jones if g1j else None,
            da_base_vis if bvis else None,
            da_g2_jones if g2j else None)

    stream_model_vis = predict_vis(*args, streams=streams)
    fan_model_vis = predict_vis(*args, streams=None)

    stream_model_vis, fan_model_vis = dask.compute(stream_model_vis,
                                                   fan_model_vis)

    assert_array_almost_equal(fan_model_vis, np_model_vis)
    assert_array_almost_equal(stream_model_vis, fan_model_vis)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import numpy as np

from africanus.util.docs import DocstringTemplate
from africanus.util.numba import jit


@jit(nopython=True, nogil=True, cache=True)
def _fit_spi_components_impl(data, weights, freqs, freq0, out,
                             jac, ncomps, nfreqs, tol, maxiter):
    w = freqs/freq0
    for comp in range(ncomps):
        eps = 1.0
        k = 0
        alphak = out[0, comp]
        i0k = out[2, comp]
        while eps > tol and k < maxiter:
            alphap = alphak
            i0p = i0k
            jac[1, :] = w**alphak
            model = i0k*jac[1, :]
            jac[0, :] = model * np.log(w)
            residual = data[comp] - model
            lik = 0.0
            hess00 = 0.0
            hess01 = 0.0
            hess11 = 0.0
            jr0 = 0.0
            jr1 = 0.0
            for v in range(nfreqs):
                lik += residual[v] * weights[v] * residual[v]
                jr0 += jac[0, v] * weights[v] * residual[v]
                jr1 += jac[1, v] * weights[v] * residual[v]
                hess00 += jac[0, v] * weights[v] * jac[0, v]
                hess01 += jac[0, v] * weights[v] * jac[1, v]
                hess11 += jac[1, v] * weights[v] * jac[1, v]
            det = hess00 * hess11 - hess01**2
            alphak = alphap + (hess11 * jr0 - hess01 * jr1)/det
            i0k = i0p + (-hess01 * jr0 + hess00 * jr1)/det
            eps = np.maximum(np.abs(alphak - alphap), np.abs(i0k - i0p))
            k += 1
        if k == maxiter:
            print("Warning - max iterations exceeded for component ", comp)
        out[0, comp] = alphak
        out[1, comp] = hess11/det
        out[2, comp] = i0k
        out[3, comp] = hess00/det
    return out


def fit_spi_components(data, weights, freqs, freq0,
                       alphai=None, I0i=None, tol=1e-4,
                       maxiter=100):
    ncomps, nfreqs = data.shape
    jac = np.zeros((2, nfreqs), dtype=data.dtype)
    out = np.zeros((4, ncomps), dtype=data.dtype)
    if alphai is not None:
        out[0, :] = alphai
    else:
        out[0, :] = -0.7 * np.ones(ncomps, dtype=data.dtype)
    if I0i is not None:
        out[2, :] = I0i
    else:
        tmp = np.abs(freqs - freq0)
        ref_freq_idx = np.argwhere(tmp == tmp.min()).squeeze()
        out[2, :] = data[:, ref_freq_idx]
    return _fit_spi_components_impl(data, weights, freqs, freq0, out,
                                    jac, ncomps, nfreqs, tol, maxiter)


SPI_DOCSTRING = DocstringTemplate(
    r"""
    Computes the spectral indices and the intensity
    at the reference frequency of a spectral index model:

    .. math::

        I(\nu) = I(\nu_0) \left( \frac{\nu}{\nu_0} \right) ^ \alpha

    Parameters
    ----------

    data : $(array_type)
        array of shape :code:`(comps, chan)`
        The noisy data as a function of frequency.
    weights : $(array_type)
        array of shape :code:`(chan,)`
        Inverse of variance on each frequency axis.
    freqs : $(array_type)
        frequencies of shape :code:`(chan,)`
    freq0 : float
        Reference frequency
    alphai : $(array_type), optional
        array of shape :code:`(comps,)`
        Initial guess for the alphas. Defaults
        to -0.7.
    I0i : $(array_type), optional
        array of shape :code:`(comps,)`
        Initial guess for the intensities at the
        reference frequency. Defaults to 1.0.
    tol : float, optional
        Solver absolute tolerance (optional).
        Defaults to 1e-6.
    maxiter : int, optional
        Solver maximum iterations (optional).
        Defaults to 100.
    dtype : np.dtype, optional
        Datatype of result. Should be either np.float32
        or np.float64. Defaults to np.float64.

    Returns
    -------
    out : $(array_type)
        array of shape :code:`(4, comps)`
        The fitted components arranged
        as [alphas, alphavars, I0s, I0vars]
    """)

try:
    fit_spi_components.__doc__ = SPI_DOCSTRING.substitute(
                            array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



__all__ = ["c", "minus_two_pi_over_c", "two_pi_over_c"]

import numpy as np

from ..util.docs import on_rtd

# Lightspeed
c = 2.99792458e8

two_pi_over_c = 2 * 3.14 / c if on_rtd() else 2 * np.pi / c
minus_two_pi_over_c = -two_pi_over_c

DEG2RAD = 3.14 / 180.0 if on_rtd() else np.pi / 180.0
ARCSEC2RAD = 3.14 / (180 * 3600) if on_rtd() else np.pi / (180 * 3600)



# -*- coding: utf-8 -*-


import numpy as np

from africanus.model.wsclean.file_model import load


def test_wsclean_model_file(wsclean_model_file):
    sources = dict(load(wsclean_model_file))

    (name, stype, ra, dec, I,
     spi, log_si, ref_freq) = (sources[n] for n in (
                                            "Name", "Type", "Ra", "Dec", "I",
                                            "SpectralIndex", "LogarithmicSI",
                                            "ReferenceFrequency"))

    # Seven sources
    assert (len(I) == len(spi) == len(log_si) == len(ref_freq) == 7)

    # Name and type read correctly
    assert name[0] == "s0c0" and stype[0] == "POINT"

    # Check ra conversion for line 0 file entry (-float, float, float)
    hours, mins, secs = (-8., 28., 5.152)
    expected_ra0 = -2.0 * np.pi * (
                    (-hours / 24.0) +
                    (mins / (24.0*60.0)) +
                    (secs / (24.0*60.0*60.0)))

    assert ra[0] == expected_ra0

    # Check dec conversion for line 0 file entry
    degs, mins, secs = (39., 35., 8.511)
    expected_dec0 = 2.0 * np.pi * (
                     (degs / 360.0) +
                     (mins / (360.0*60.0)) +
                     (secs / (360.0*60.0*60.0)))

    assert dec[0] == expected_dec0

    # SPI read correctly
    assert spi[0] == [-0.00695379313004673, -0.0849693907803257]

    # LogrithmicSI read correctly
    assert log_si[0] is True

    # Check ra conversion for line 2 file entry (int, not float, seconds)
    hours, mins, secs = (8, 18, 44)
    expected_ra2 = 2.0 * np.pi * (
                    (hours / 24.0) +
                    (mins / (24.0*60.0)) +
                    (secs / (24.0*60.0*60.0)))

    assert ra[2] == expected_ra2

    # Check dec conversion for line 2 file entry (int, not float, seconds)
    degs, mins, secs = (39, 38, 37)
    expected_dec2 = 2.0 * np.pi * (
                     (degs / 360.0) +
                     (mins / (360.0*60.0)) +
                     (secs / (360.0*60.0*60.0)))

    assert dec[2] == expected_dec2

    assert log_si[2] is False

    assert I[2] == 0.000233552686127518

    # Check dec conversion for line 4 file entry (+int, not float, seconds)
    degs, mins, secs = (+41, 47, 17.131)
    expected_dec4 = 2.0 * np.pi * (
                     (degs / 360.0) +
                     (mins / (360.0*60.0)) +
                     (secs / (360.0*60.0*60.0)))

    assert dec[4] == expected_dec4

    # Missing reference frequency set in the last
    assert ref_freq[-1] == ref_freq[0]

    # Last name and type correct
    assert name[-1] == "s1c2" and stype[-1] == "GAUSSIAN"

    assert I[-1] == 0.000660490865128381



# flake8: noqa

from africanus.rime.cuda.beam import beam_cube_dde
from africanus.rime.cuda.feeds import feed_rotation
from africanus.rime.cuda.phase import phase_delay
from africanus.rime.cuda.predict import predict_vis



# -*- coding: utf-8 -*-


import numpy as onp
import pytest

from africanus.rime.phase import phase_delay as np_phase_delay
from africanus.rime.jax.phase import phase_delay


@pytest.mark.parametrize("dtype", [onp.float32, onp.float64])
def test_jax_phase_delay(dtype):
    jax = pytest.importorskip('jax')
    np = pytest.importorskip('jax.numpy')

    onp.random.seed(0)

    uvw = onp.random.random(size=(100, 3)).astype(dtype)
    lm = onp.random.random(size=(10, 2)).astype(dtype)*0.001
    frequency = onp.linspace(.856e9, .856e9*2, 64).astype(dtype)

    # Compute complex phase
    np_complex_phase = np_phase_delay(lm, uvw, frequency)
    complex_phase = jax.jit(phase_delay)(lm, uvw, frequency)

    onp.testing.assert_array_almost_equal(complex_phase, np_complex_phase)
    expected_ctype = np.result_type(dtype, np.complex64)
    assert np_complex_phase.dtype == expected_ctype
    assert complex_phase.dtype == expected_ctype



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

"""The setup script."""

import os
from setuptools import setup, find_packages

# Import requirements
from africanus.install.requirements import (requirements,
                                            extras_require,
                                            setup_requirements,
                                            test_requirements)

with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    author="Simon Perkins",
    author_email='sperkins@ska.ac.za',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    description="Radio Astronomy Building Blocks",
    entry_points={
        'console_scripts': [
            'plot-filter=africanus.filters.plot_filter:main',
            'plot-taper=africanus.filters.plot_taper:main'],
    },
    extras_require=extras_require,
    install_requires=requirements,
    license="GNU General Public License v2",
    long_description=readme,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='codex-africanus',
    name='codex-africanus',
    packages=find_packages(),
    python_requires=">=3.5",
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/ska-sa/codex-africanus',
    version='0.2.0',
    zip_safe=False,
)



import numpy as np
import pytest

from africanus.rime.parangles import _discovered_backends

no_casa = 'casa' not in _discovered_backends
no_astropy = 'astropy' not in _discovered_backends


def _julian_day(year, month, day):
    """
    Given a Anno Dominei date, computes the Julian Date in days.

    Parameters
    ----------
    year : int
    month : int
    day : int

    Returns
    -------
    float
        Julian Date
    """

    # Formula below from
    # http://scienceworld.wolfram.com/astronomy/JulianDate.html
    # Also agrees with https://gist.github.com/jiffyclub/1294443
    return (367*year - int(7*(year + int((month+9)/12))/4)
            - int((3*(int(year + (month - 9)/7)/100)+1)/4)
            + int(275*month/9) + day + 1721028.5)


def _modified_julian_date(year, month, day):
    """
    Given a Anno Dominei date, computes the Modified Julian Date in days.

    Parameters
    ----------
    year : int
    month : int
    day : int

    Returns
    -------
    float
        Modified Julian Date
    """

    return _julian_day(year, month, day) - 2400000.5


def _observation_endpoints(year, month, day, hour_duration):
    """
    Start and end points of an observation starting on
    ``year-month-day`` and of duration ``hour_duration``
    in Modified Julian Date seconds
    """
    start = _modified_julian_date(year, month, day)
    end = start + hour_duration / 24.

    # Convert to seconds
    start *= 86400.
    end *= 86400.

    return (start, end)


@pytest.mark.flaky(min_passes=1, max_runs=3)
@pytest.mark.parametrize('backend', [
    'test',
    pytest.param('casa', marks=pytest.mark.skipif(
                    no_casa,
                    reason='python-casascore not installed')),
    pytest.param('astropy', marks=pytest.mark.skipif(
                    no_astropy,
                    reason="astropy not installed"))])
@pytest.mark.parametrize('observation', [(2018, 1, 1, 4)])
def test_parallactic_angles(observation, wsrt_ants, backend):
    import numpy as np
    from africanus.rime import parallactic_angles

    start, end = _observation_endpoints(*observation)
    time = np.linspace(start, end, 5)
    ant = wsrt_ants[:4, :]
    fc = np.random.random((2,)).astype(np.float64)

    pa = parallactic_angles(time, ant, fc, backend=backend)
    assert pa.shape == (5, 4)


@pytest.mark.flaky(min_passes=1, max_runs=3)
@pytest.mark.skipif(no_casa or no_astropy,
                    reason="Neither python-casacore or astropy installed")
# Parametrize on observation length and error tolerance
@pytest.mark.parametrize('obs_and_tol', [
    ((2018, 1, 1, 4), "10s"),
    ((2018, 2, 20, 8), "10s"),
    ((2018, 11, 2, 4), "10s")])
def test_compare_astropy_and_casa(obs_and_tol, wsrt_ants):
    """
    Compare astropy and python-casacore parallactic angle implementations.
    More work needs to be done here to get things lined up closer,
    but the tolerances above suggest nothing > 10 arcseconds.
    """
    import numpy as np
    from africanus.rime.parangles_casa import casa_parallactic_angles
    from africanus.rime.parangles_astropy import astropy_parallactic_angles
    from astropy import units
    from astropy.coordinates import Angle

    obs, rtol = obs_and_tol
    start, end = _observation_endpoints(*obs)

    time = np.linspace(start, end, 5)
    ant = wsrt_ants[:4, :]
    fc = np.array([0., 1.04719755], dtype=np.float64)

    astro_pa = astropy_parallactic_angles(time, ant, fc)
    casa_pa = casa_parallactic_angles(time, ant, fc, zenith_frame='AZELGEO')

    # Convert to angle degrees
    astro_pa = Angle(astro_pa, unit=units.deg).wrap_at(180*units.deg)
    casa_pa = Angle(casa_pa*units.rad, unit=units.deg).wrap_at(180*units.deg)

    # Difference in degrees, wrapped at 180
    diff = np.abs((astro_pa - casa_pa).wrap_at(180*units.deg))
    assert np.all(np.abs(diff) < Angle(rtol))


@pytest.mark.flaky(min_passes=1, max_runs=3)
@pytest.mark.parametrize('backend', [
    'test',
    pytest.param('casa', marks=pytest.mark.skipif(
                                no_casa,
                                reason='python-casascore not installed')),
    pytest.param('astropy', marks=pytest.mark.skipif(
                                no_astropy,
                                reason="astropy not installed"))])
@pytest.mark.parametrize('observation', [(2018, 1, 1, 4)])
def test_dask_parallactic_angles(observation, wsrt_ants, backend):
    da = pytest.importorskip('dask.array')
    from africanus.rime import parallactic_angles as np_parangle
    from africanus.rime.dask import parallactic_angles as da_parangle

    start, end = _observation_endpoints(*observation)
    np_times = np.linspace(start, end, 5)
    np_ants = wsrt_ants[:4, :]
    np_fc = np.random.random(size=2)

    np_pa = np_parangle(np_times, np_ants, np_fc, backend=backend)
    np_pa = np.asarray(np_pa)

    da_times = da.from_array(np_times, chunks=(2, 3))
    da_ants = da.from_array(np_ants, chunks=((2, 2), 3))
    da_fc = da.from_array(np_fc, chunks=2)

    da_pa = da_parangle(da_times, da_ants, da_fc, backend=backend)

    assert np.all(np_pa == da_pa.compute())



# -*- coding: utf-8 -*-


import os
import re
from string import Template

_on_rtd = bool(os.environ.get("READTHEDOCS"))


def on_rtd():
    return _on_rtd


def mod_docs(docstring, replacements):
    for search, replace in replacements:
        docstring = docstring.replace(search, replace)

    return docstring


def doc_tuple_to_str(doc_tuple, replacements=None):
    fields = getattr(doc_tuple, "_fields", None)

    if fields is not None:
        fields = (getattr(doc_tuple, f) for f in doc_tuple._fields)
    elif isinstance(doc_tuple, dict):
        fields = doc_tuple.values()

    if replacements is not None:
        fields = (mod_docs(f, replacements) for f in fields)

    return ''.join(fields)


class DocstringTemplate(Template):
    """
    Overrides the ${identifer} braced pattern in the string Template
    with a $(identifier) braced pattern
    """
    pattern = r"""
    %(delim)s(?:
      (?P<escaped>%(delim)s)   |   # Escape sequence of two delimiters
      (?P<named>%(id)s)        |   # delimiter and a Python identifier
      \((?P<braced>%(id)s)\)   |   # delimiter and a braced identifier
      (?P<invalid>)                # Other ill-formed delimiter exprs
    )
    """ % {'delim': re.escape(Template.delimiter),
           'id': Template.idpattern}



# -*- coding: utf-8 -*-


import numpy as np
import numba

from africanus.util.numba import generated_jit, njit


@njit(nogil=True, cache=True)
def _unique_internal(data):
    if len(data.shape) != 1:
        raise ValueError("_unique_internal currently "
                         "only supports 1D arrays")

    # Handle the empty array case
    if data.shape[0] == 0:
        return (data,
                np.empty((0,), dtype=np.intp),
                np.empty((0,), dtype=np.intp),
                np.empty((0,), dtype=np.intp))

    # See numpy's unique1d
    perm = np.argsort(data, kind='mergesort')

    # Combine these arrays to save on allocations?
    aux = np.empty_like(data)
    mask = np.empty(aux.shape, dtype=np.bool_)
    inv_idx = np.empty(mask.shape, dtype=np.intp)

    # Hard code first iteration
    p = perm[0]
    aux[0] = data[p]
    mask[0] = True
    cumsum = 1
    inv_idx[p] = cumsum - 1
    counts = [np.intp(0)]

    for i in range(1, aux.shape[0]):
        p = perm[i]
        aux[i] = data[p]
        d = aux[i] != aux[i - 1]
        mask[i] = d
        cumsum += d
        inv_idx[p] = cumsum - 1

        if d:
            counts.append(np.intp(i))

    counts.append(aux.shape[0])

    # (uniques, indices, inverse index, counts)
    return aux[mask], perm[mask], inv_idx, np.diff(np.array(counts))


@generated_jit(nopython=True, nogil=True, cache=True)
def unique_time(time):
    """ Return unique time, inverse index and counts """
    if time.dtype not in (numba.float32, numba.float64):
        raise ValueError("time must be floating point but is %s" % time.dtype)

    def impl(time):
        return _unique_internal(time)

    return impl


@generated_jit(nopython=True, nogil=True, cache=True)
def unique_baselines(ant1, ant2):
    """ Return unique baselines, inverse index and counts """
    if not ant1.dtype == numba.int32 or not ant2.dtype == numba.int32:
        # Need these to be int32 for the bl_32bit.view(np.int64) trick
        raise ValueError("ant1 and ant2 must be np.int32 "
                         "but received %s and %s" %
                         (ant1.dtype, ant2.dtype))

    def impl(ant1, ant2):
        # Trickery, stack the two int32 antenna pairs in an array
        # and cast to int64
        bl_32bit = np.empty((ant1.shape[0], 2), dtype=np.int32)

        # Copy data
        for r in range(ant1.shape[0]):
            bl_32bit[r, 0] = ant1[r]
            bl_32bit[r, 1] = ant2[r]

        # Cast to int64 for the unique operation
        bl = bl_32bit.view(np.int64).reshape(ant1.shape[0])

        ret, idx, inv, counts = _unique_internal(bl)

        # Recast to int32 and reshape
        ubl = ret.view(np.int32).reshape(ret.shape[0], 2)

        return ubl, idx, inv, counts

    return impl



# -*- coding: utf-8 -*-


def aggregate_chunks(chunks, max_chunks):
    """
    Aggregate dask ``chunks`` together into chunks no larger than
    ``max_chunks``.

    .. code-block:: python

        chunks, max_c = ((3,4,6,3,6,7),(1,1,1,1,1,1)), (10,3)
        expected = ((7,9,6,7), (2,2,1,1))
        assert aggregate_chunks(chunks, max_c) == expected


    Parameters
    ----------
    chunks : sequence of tuples or tuple
    max_chunks : sequence of ints or int

    Returns
    -------
    sequence of tuples or tuple

    """

    if isinstance(max_chunks, int):
        chunks = (chunks,)
        max_chunks = (max_chunks,)

    singleton = True if len(max_chunks) == 1 else False

    if len(chunks) != len(max_chunks):
        raise ValueError("len(chunks) != len(max_chunks)")

    if not all(len(chunks[0]) == len(c) for c in chunks):
        raise ValueError("Number of chunks do not match")

    agg_chunks = [[] for _ in max_chunks]
    agg_chunk_counts = [0] * len(max_chunks)
    chunk_scratch = [0] * len(max_chunks)
    ndim = len(chunks[0])

    # For each chunk dimension
    for di in range(ndim):
        # For each chunk
        aggregate = False

        for ci, chunk in enumerate(chunks):
            chunk_scratch[ci] = agg_chunk_counts[ci] + chunk[di]
            if chunk_scratch[ci] > max_chunks[ci]:
                aggregate = True

        if aggregate:
            for ci, chunk in enumerate(chunks):
                agg_chunks[ci].append(agg_chunk_counts[ci])
                agg_chunk_counts[ci] = chunk[di]
        else:
            for ci, chunk in enumerate(chunks):
                agg_chunk_counts[ci] = chunk_scratch[ci]

    # Do the final aggregation
    for ci, chunk in enumerate(chunks):
        agg_chunks[ci].append(agg_chunk_counts[ci])
        agg_chunk_counts[ci] = chunk[di]

    agg_chunks = tuple(tuple(ac) for ac in agg_chunks)

    return agg_chunks[0] if singleton else agg_chunks


def corr_shape(ncorr, corr_shape):
    """
    Returns the shape of the correlations, given
    ``ncorr`` and the type of correlation shape requested

    Parameters
    ----------
    ncorr : integer
        Number of correlations
    corr_shape : {'flat', 'matrix'}
        Shape of output correlations


    Returns
    -------
    tuple
        Shape tuple describing the correlation dimensions

        * If ``flat`` returns :code:`(ncorr,)`
        * If ``matrix`` returns

            * :code:`(1,)` if :code:`ncorr == 1`
            * :code:`(2,)` if :code:`ncorr == 2`
            * :code:`(2,2)` if :code:`ncorr == 4`


    """
    if corr_shape == "flat":
        return (ncorr,)
    elif corr_shape == "matrix":
        if ncorr == 1:
            return (1,)
        elif ncorr == 2:
            return (2,)
        elif ncorr == 4:
            return (2, 2)
        else:
            raise ValueError("ncorr not in (1, 2, 4)")
    else:
        raise ValueError("corr_shape must be 'flat' or 'matrix'")



# -*- coding: utf-8 -*-


from collections import OrderedDict
import string
import re

import numpy as np

from africanus.util.casa_types import STOKES_ID_MAP


class FitsAxes(object):
    """
    FitsAxes object, inspired by Tigger's FITSAxes
    """

    def __init__(self, header=None):
        # Create an zero-dimensional object if no header supplied
        self._ndims = ndims = 0 if header is None else header['NAXIS']

        # Extract header information for each dimension
        axr = list(range(1, ndims+1))
        self._naxis = [header.get('NAXIS%d' % n) for n in axr]
        self._ctype = [header.get('CTYPE%d' % n, n).strip() for n in axr]
        self._crval = [header.get('CRVAL%d' % n, 0) for n in axr]
        # Convert right pixel from FORTRAN to C indexing
        self._crpix = [header['CRPIX%d' % n]-1 for n in axr]
        self._cdelt = [header.get('CDELT%d' % n, 1) for n in axr]
        self._cunit = [header.get('CUNIT%d' % n, '').strip().upper()
                       for n in axr]


class BeamAxes(FitsAxes):
    """
    Describes the FITS axes of a BEAM cube.

    In general, FORTRAN ordered indices are
    converted to C ordering
    (CRPIX and the individual axis indices)

    Any degree axes are converted to radians and
    grids for each axis are created.

    Inversions of the L, M, X and Y grids are supported
    if a minus sign is detected before the CUNIT in
    the FITS header.
    """

    def __init__(self, header=None):
        super(BeamAxes, self).__init__(header)

        # Check for custom irregular grid format.
        # Currently only implemented for FREQ dimension.
        irregular_grid = np.asarray([
            [header.get('G%s%d' % (self._ctype[i], j), None)
             for j in range(1, self._naxis[i]+1)]
            for i in range(self._ndims)])

        # Irregular grids are only valid if values exist for all grid points
        self._irreg = [all(x is not None for x in irregular_grid[i])
                       for i in range(self._ndims)]

        def _regular_grid(i):
            """ Construct a regular grid from a FitsAxes object and index """
            R = np.arange(0.0, float(self._naxis[i]))
            return (R - self._crpix[i])*self._cdelt[i] + self._crval[i]

        # Set up the grid
        self._grid = [_regular_grid(i) if not self._irreg[i]
                      else np.asarray(irregular_grid[i])
                      for i in range(self._ndims)]

        self._sign = [1.0]*self._ndims

        for i in range(self._ndims):
            # Convert any degree axes to radians
            if self._cunit[i] == 'DEG':
                self._cunit[i] = 'RAD'
                self._crval[i] = np.deg2rad(self._crval[i])
                self._cdelt[i] = np.deg2rad(self._cdelt[i])
                self._grid[i] = np.deg2rad(self._grid[i])

            # Flip the sign and correct the ctype if necessary
            if self._ctype[i].startswith('-'):
                self._ctype[i] = self._ctype[i][1]
                self._sign[i] = -1.0

    @property
    def ndims(self):
        return self._ndims

    @property
    def crpix(self):
        return self._crpix

    @property
    def naxis(self):
        return self._naxis

    @property
    def crval(self):
        return self._crval

    @property
    def cdelt(self):
        return self._cdelt

    @property
    def cunit(self):
        return self._cunit

    @property
    def ctype(self):
        return self._ctype

    @property
    def grid(self):
        return self._grid

    @property
    def sign(self):
        return self._sign


def beam_grids(header):
    """
    Extracts the FITS indices and grids for the beam dimensions
    in the supplied FITS ``header``.
    Specifically the axes specified by

    1. ``L`` or ``X`` CTYPE
    2. ``M`` or ``Y`` CTYPE
    3. ``FREQ`` CTYPE

    If the first two axes have a negative sign, such as ``-L``, the grid
    will be inverted.

    Any grids corresponding to axes with a CUNIT type of ``DEG``
    will be converted to radians.

    Parameters
    ----------
    header : :class:`~astropy.io.fits.Header` or dict
        FITS header object.

    Returns
    -------
    tuple
        Returns
        ((l_axis, l_grid), (m_axis, m_grid), (freq_axis, freq_grid))
        where the axis is the FORTRAN indexed FITS axis (1-indexed)
        and grid contains the values at each pixel along the axis.
    """
    beam_axes = BeamAxes(header)

    l = m = freq = None  # noqa

    # Find the relevant axes
    for i in range(beam_axes.ndims):
        if beam_axes.ctype[i] in ('L', 'X'):
            l = i  # noqa
        elif beam_axes.ctype[i] in ('M', 'Y'):
            m = i
        elif beam_axes.ctype[i] == "FREQ":
            freq = i

    # Complain if not found
    if l is None:
        raise ValueError("No L/X axis present in FITS header")

    if m is None:
        raise ValueError("No M/Y axis present in FITS header")

    if freq is None:
        raise ValueError("No FREQ axis present in FITS header")

    # Sign of L/M axes?
    l_sign = beam_axes.sign[l]
    m_sign = beam_axes.sign[m]

    # Obtain axes grids
    l_grid = beam_axes.grid[l]
    m_grid = beam_axes.grid[m]
    freq_grid = beam_axes.grid[freq]

    # flip the grid around if signs are different
    l_grid = np.flipud(l_grid) if l_sign == -1.0 else l_grid
    m_grid = np.flipud(m_grid) if m_sign == -1.0 else m_grid

    return ((l+1, l_grid), (m+1, m_grid), (freq+1, freq_grid))


class FitsFilenameTemplate(string.Template):
    """
    Overrides the ${identifer} braced pattern in the string Template
    with a $(identifier) braced pattern expected by FITS beam filename
    schema
    """
    pattern = r"""
    %(delim)s(?:
      (?P<escaped>%(delim)s)   |   # Escape sequence of two delimiters
      (?P<named>%(id)s)        |   # delimiter and a Python identifier
      \((?P<braced>%(id)s)\)   |   # delimiter and a braced identifier
      (?P<invalid>)                # Other ill-formed delimiter exprs
    )
    """ % {'delim': re.escape(string.Template.delimiter),
           'id': string.Template.idpattern}


CIRCULAR_CORRELATIONS = ('rr', 'rl', 'lr', 'll')
LINEAR_CORRELATIONS = ('xx', 'xy', 'yx', 'yy')
REIM = ('re', 'im')


def _re_im_filenames(corr, template):
    filenames = []

    for ri in REIM:
        try:
            filename = template.substitute(corr=corr.lower(),
                                           CORR=corr.upper(),
                                           reim=ri.lower(),
                                           REIM=ri.upper())
        except KeyError:
            raise ValueError("Invalid filename schema '%s'. "
                             "FITS Beam filename schemas "
                             "must follow forms such as "
                             "'beam_$(corr)_$(reim).fits' or "
                             "'beam_$(CORR)_$(REIM).fits."
                             % template.template)
        else:
            filenames.append(filename)

    return filenames


def beam_filenames(filename_schema, corr_types):
    """
    Returns a dictionary of beam filename pairs,
    keyed on correlation,from the cartesian product
    of correlations and real, imaginary pairs

    Given ``beam_$(corr)_$(reim).fits`` returns:

    .. code-block:: python

        {
          'xx' : ['beam_xx_re.fits', 'beam_xx_im.fits'],
          'xy' : ['beam_xy_re.fits', 'beam_xy_im.fits'],
          ...
          'yy' : ['beam_yy_re.fits', 'beam_yy_im.fits'],
        }

    Given ``beam_$(CORR)_$(REIM).fits`` returns:

    .. code-block:: python

        {
          'xx' : ['beam_XX_RE.fits', 'beam_XX_IM.fits'],
          'xy' : ['beam_XY_RE.fits', 'beam_XY_IM.fits'],
          ...
          'yy' : ['beam_YY_RE.fits', 'beam_YY_IM.fits']),
        }

    Parameters
    ----------
    filename_schema : str
        String containing the filename schema.
    corr_types : list of integers
        list of integers defining the correlation type.

    Returns
    -------
    dict
        Dictionary of schema ``{correlation : (refile, imfile)}``
        mapping correlations to real and imaginary filename pairs

    """
    template = FitsFilenameTemplate(filename_schema)

    corr_names = []

    for corr_type in corr_types:
        try:
            corr_name = STOKES_ID_MAP[corr_type]
        except KeyError:
            raise ValueError("Unknown Stokes ID %d" % corr_type)
        else:
            corr_names.append(corr_name.lower())

    return OrderedDict((c, _re_im_filenames(c, template))
                       for c in corr_names)



# -*- coding: utf-8 -*-

from functools import reduce

import numpy as np
from africanus.util.docs import DocstringTemplate
from africanus.util.numba import njit


@njit(nogil=True, cache=True)
def freq_grid_interp(frequency, beam_freq_map):
    # Interpolated grid coordinate
    beam_nud = beam_freq_map.shape[0]
    freq_data = np.empty((frequency.shape[0], 3),
                         dtype=frequency.dtype)

    for f in range(frequency.shape[0]):
        freq = frequency[f]
        lower = 0
        upper = beam_nud - 1

        while lower <= upper:
            mid = lower + (upper - lower) // 2
            beam_freq = beam_freq_map[mid]

            if beam_freq < freq:
                lower = mid + 1
            elif beam_freq > freq:
                upper = mid - 1
            else:
                lower = mid
                break

        # This handles the lower <= upper in the while loop
        lower = min(lower, upper)
        upper = lower + 1

        # Set up scaling, lower weight, lower grid pos
        if lower == -1:
            freq_data[f, 0] = freq / beam_freq_map[0]
            freq_data[f, 1] = 1.0
            freq_data[f, 2] = 0
        elif upper == beam_nud:
            freq_data[f, 0] = freq / beam_freq_map[beam_nud - 1]
            freq_data[f, 1] = 0.0
            freq_data[f, 2] = beam_nud - 2
        else:
            freq_data[f, 0] = 1.0
            freq_low = beam_freq_map[lower]
            freq_high = beam_freq_map[upper]
            freq_diff = freq_high - freq_low
            freq_data[f, 1] = (freq_high - freq) / freq_diff
            freq_data[f, 2] = lower

    return freq_data


@njit(nogil=True, cache=True)
def beam_cube_dde(beam, beam_lm_extents, beam_freq_map,
                  lm, parallactic_angles, point_errors, antenna_scaling,
                  frequency):

    nsrc = lm.shape[0]
    ntime, nants = parallactic_angles.shape
    nchan = frequency.shape[0]
    beam_lw, beam_mh, beam_nud = beam.shape[:3]
    corrs = beam.shape[3:]

    if beam_lw < 2 or beam_mh < 2 or beam_nud < 2:
        raise ValueError("beam_lw, beam_mh and beam_nud must be >= 2")

    # Flatten correlations
    ncorrs = reduce(lambda x, y: x*y, corrs, 1)

    lower_l, upper_l = beam_lm_extents[0]
    lower_m, upper_m = beam_lm_extents[1]

    ex_dtype = beam_lm_extents.dtype

    # Maximum l and m indices in float and int
    lmaxf = ex_dtype.type(beam_lw - 1)
    mmaxf = ex_dtype.type(beam_mh - 1)
    lmaxi = beam_lw - 1
    mmaxi = beam_mh - 1

    lscale = lmaxf / (upper_l - lower_l)
    mscale = mmaxf / (upper_m - lower_m)

    one = ex_dtype.type(1)
    zero = ex_dtype.type(0)

    # Flatten the beam on correlation
    fbeam = beam.reshape((beam_lw, beam_mh, beam_nud, ncorrs))

    # Allocate output array with correlations flattened
    fjones = np.empty((nsrc, ntime, nants, nchan, ncorrs), dtype=beam.dtype)

    # Compute frequency interpolation stuff
    freq_data = freq_grid_interp(frequency, beam_freq_map)

    corr_sum = np.zeros((ncorrs,), dtype=beam.dtype)
    absc_sum = np.zeros((ncorrs,), dtype=beam.real.dtype)
    beam_scratch = np.zeros((ncorrs,), dtype=beam.dtype)

    for t in range(ntime):
        for a in range(nants):
            sin_pa = np.sin(parallactic_angles[t, a])
            cos_pa = np.cos(parallactic_angles[t, a])

            for s in range(nsrc):
                # Extract lm coordinates
                l, m = lm[s]

                for f in range(nchan):
                    # Unpack frequency data
                    freq_scale = freq_data[f, 0]
                    # lower and upper frequency weights
                    nud = freq_data[f, 1]
                    inv_nud = 1.0 - nud
                    # lower and upper frequency grid position
                    gc0 = np.int32(freq_data[f, 2])
                    gc1 = gc0 + 1

                    # Apply any frequency scaling
                    sl = l * freq_scale
                    sm = m * freq_scale

                    # Add pointing errors
                    tl = sl + point_errors[t, a, f, 0]
                    tm = sm + point_errors[t, a, f, 1]

                    # Rotate lm coordinate angle
                    vl = tl*cos_pa - tm*sin_pa
                    vm = tl*sin_pa + tm*cos_pa

                    # Scale by antenna scaling
                    vl *= antenna_scaling[a, f, 0]
                    vm *= antenna_scaling[a, f, 1]

                    # Shift into the cube coordinate system
                    vl = lscale*(vl - lower_l)
                    vm = mscale*(vm - lower_m)

                    # Clamp the coordinates to the edges of the cube
                    vl = max(zero, min(vl, lmaxf))
                    vm = max(zero, min(vm, mmaxf))

                    # Snap to the lower grid coordinates
                    gl0 = np.int32(np.floor(vl))
                    gm0 = np.int32(np.floor(vm))

                    # Snap to the upper grid coordinates
                    gl1 = min(gl0 + 1, lmaxi)
                    gm1 = min(gm0 + 1, mmaxi)

                    # Difference between grid and offset coordinates
                    ld = vl - gl0
                    md = vm - gm0

                    # Zero accumulation arrays
                    corr_sum[:] = 0
                    absc_sum[:] = 0

                    # Accumulate lower cube correlations
                    beam_scratch[:] = fbeam[gl0, gm0, gc0, :]
                    weight = (one - ld)*(one - md)*nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    beam_scratch[:] = fbeam[gl1, gm0, gc0, :]
                    weight = ld*(one - md)*nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    beam_scratch[:] = fbeam[gl0, gm1, gc0, :]
                    weight = (one - ld)*md*nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    beam_scratch[:] = fbeam[gl1, gm1, gc0, :]
                    weight = ld*md*nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    # Accumulate upper cube correlations
                    beam_scratch[:] = fbeam[gl0, gm0, gc1, :]
                    weight = (one - ld)*(one - md)*inv_nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    beam_scratch[:] = fbeam[gl1, gm0, gc1, :]
                    weight = ld*(one - md)*inv_nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    beam_scratch[:] = fbeam[gl0, gm1, gc1, :]
                    weight = (one - ld)*md*inv_nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    beam_scratch[:] = fbeam[gl1, gm1, gc1, :]
                    weight = ld*md*inv_nud

                    for c in range(ncorrs):
                        absc_sum[c] += weight * np.abs(beam_scratch[c])
                        corr_sum[c] += weight * beam_scratch[c]

                    for c in range(ncorrs):
                        # Added all correlations, normalise
                        div = np.abs(corr_sum[c])

                        if div == 0.0:
                            # This case probably works out to a zero assign
                            corr_sum[c] *= absc_sum[c]
                        else:
                            corr_sum[c] *= absc_sum[c] / div

                    # Assign normalised values
                    fjones[s, t, a, f, :] = corr_sum

    return fjones.reshape((nsrc, ntime, nants, nchan) + corrs)


BEAM_CUBE_DOCS = DocstringTemplate(
    r"""
    Evaluates Direction Dependent Effects along a source's path
    by interpolating the values  of a complex beam cube
    at the source location.

    Notes
    -----
    1. Sources are clamped to the provided `beam_lm_extents`.
    2. Frequencies outside the cube (i.e. outside beam_freq_map)
       introduce linear scaling to the lm coordinates of a source.

    Parameters
    ----------
    beam : $(array_type)
        Complex beam cube of
        shape :code:`(beam_lw, beam_mh, beam_nud, corr, corr)`.
        `beam_lw`, `beam_mh` and `beam_nud` define the size
        of the cube in the l, m and frequency dimensions, respectively.
    beam_lm_extents : $(array_type)
        lm extents of the beam cube of shape :code:`(2, 2)`.
        ``[[lower_l, upper_l], [lower_m, upper_m]]``.
    beam_freq_map : $(array_type)
        Beam frequency map of shape :code:`(beam_nud,)`.
        This array is used to define interpolation along
        the :code:`(chan,)` dimension.
    lm : $(array_type)
        Source lm coordinates of shape :code:`(source, 2)`.
        These coordinates are:

            1. Scaled if the associated frequency lies outside the beam cube.
            2. Offset by pointing errors: ``point_errors``
            3. Rotated by parallactic angles: ``parallactic_angles``.
            4. Scaled by antenna scaling factors: ``antenna_scaling``.

    parallactic_angles : $(array_type)
        Parallactic angles of shape :code:`(time, ant)`.
    point_errors : $(array_type)
        Pointing errors of shape :code:`(time, ant, chan, 2)`.
    antenna_scaling : $(array_type)
        Antenna scaling factors of shape :code:`(ant, chan, 2)`
    frequency : $(array_type)
        Frequencies of shape :code:`(chan,)`.

    Returns
    -------
    ddes : $(array_type)
        Direction Dependent Effects of shape
        :code:`(source, time, ant, chan, corr, corr)`
    """)


try:
    beam_cube_dde.__doc__ = BEAM_CUBE_DOCS.substitute(
                                array_type=":class:`numpy.ndarray`")
except AttributeError:
    pass



# -*- coding: utf-8 -*-


try:
    import jax.numpy as np
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None

from africanus.constants import minus_two_pi_over_c
from africanus.util.requirements import requires_optional


@requires_optional('jax', opt_import_error)
def phase_delay(lm, uvw, frequency):
    out_dtype = np.result_type(lm, uvw, frequency, np.complex64)

    one = lm.dtype.type(1.0)
    neg_two_pi_over_c = lm.dtype.type(minus_two_pi_over_c)
    complex_one = out_dtype.type(1j)

    l = lm[:, 0, None, None]  # noqa
    m = lm[:, 1, None, None]

    u = uvw[None, :, 0, None]
    v = uvw[None, :, 1, None]
    w = uvw[None, :, 2, None]

    n = np.sqrt(one - l**2 - m**2) - one

    real_phase = (neg_two_pi_over_c *
                  (l * u + m * v + n * w) *
                  frequency[None, None, :])

    return np.exp(complex_one*real_phase)



# -*- coding: utf-8 -*-


from functools import wraps

try:
    from dask.utils import SerializableLock as Lock
except ImportError:
    from threading import Lock


class SingletonMixin(object):
    __singleton_lock = Lock()
    __singleton_instance = None

    @classmethod
    def instance(cls):
        if not cls.__singleton_instance:
            with cls.__singleton_lock:
                if not cls.__singleton_instance:
                    cls.__singleton_instance = cls()

        return cls.__singleton_instance


def format_code(code):
    """
    Formats some code with line numbers

    Parameters
    ----------
    code : str
        Code

    Returns
    -------
    str
        Code prefixed with line numbers
    """
    lines = ['']
    lines.extend(["%-5d %s" % (i, l) for i, l
                  in enumerate(code.split('\n'), 1)])
    return '\n'.join(lines)


class memoize_on_key(object):
    """
    Memoize based on a key function supplied by the user.
    The key function should return a custom key
    for memoizing the decorated function, based on the arguments
    passed to it.

    In the following example, the arguments required to generate
    the `_generate_phase_delay_kernel` function are the types of
    the `lm`, `uvw` and `frequency` arrays, as well as the number
    of correlations, `ncorr`.

    The supplied ``key_fn`` produces a unique key based on these types
    and the number of correlations, which is used to cache the
    generated function.

    .. code-block:: python

        def key_fn(lm, uvw, frequency, ncorrs=4):
            '''
            Produce a unique key for the arguments of
             _generate_phase_delay_kernel
            '''
            return (lm.dtype, uvw.dtype, frequency.dtype, ncorrs)

        _code_template = jinja2.Template('''
        #define ncorrs {{ncorrs}}

        __global__ void phase_delay(
            const {{lm_type}} * lm,
            const {{uvw_type}} * uvw,
            const {{freq_type}} * frequency,
            {{out_type}} * out)
        {
            ...
        }
        ''')

        _type_map = {
            np.float32: 'float',
            np.float64: 'double'
        }

        @memoize_on_key(key_fn)
        def _generate_phase_delay_kernel(lm, uvw, frequency, ncorrs=4):
            ''' Generate the phase delay kernel '''
            out_dtype = np.result_type(lm.dtype, uvw.dtype, frequency.dtype)
            code = _code_template.render(lm_type=_type_map[lm.dtype],
                                         uvw_type=_type_map[uvw.dtype],
                                         freq_type=_type_map[frequency.dtype],
                                         ncorrs=ncorrs)
            return cp.RawKernel(code, "phase_delay")
    """

    def __init__(self, key_fn):
        self._key_fn = key_fn
        self._lock = Lock()
        self._cache = {}

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            key = self._key_fn(*args, **kwargs)

            with self._lock:
                try:
                    return self._cache[key]
                except KeyError:
                    self._cache[key] = entry = fn(*args, **kwargs)
                    return entry

        return wrapper
